# CESYS Software License - Version 1.0 - January 1st, 2016
#
# Permission is hereby granted, free of charge, to any person or organization
# obtaining a copy of the software and accompanying documentation covered by
# this license (the "Software") to use, reproduce, display, distribute,
# execute, and transmit the Software, and to prepare derivative works of the
# Software, and to permit third-parties to whom the Software is furnished to
# do so, all subject to the following:
#
# The software must only be used to operate hardware manufactured by
# CESYS GmbH, Herzogenaurach.
#
# The copyright notices in the Software and this entire statement, including
# the above license grant, this restriction and the following disclaimer,
# must be included in all copies of the Software, in whole or in part, and
# all derivative works of the Software, unless such copies or derivative
# works are solely in the form of machine-executable object code generated by
# a source language processor.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
# SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
# FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#
# Written by Thomas Hoppe <thomas.hoppe@cesys.com>, 2018

__author__  = "Thomas Hoppe"
__status__  = "production"
__version__ = "1.0"

import sys, platform
from ctypes import cdll, create_string_buffer, POINTER, byref
from ctypes import c_uint, c_int, c_char, c_char_p, c_void_p

def _getLibraryName():
    is64bit = sys.maxsize > 2**32
    baseName = "beastlink-" + __version__
    if (platform.system() == "Windows"):
        return baseName + ("-x86_64.dll" if (is64bit) else "-x86.dll")
    if (platform.system() == "Linux"):
        return "lib" + baseName + ".so"
    if (platform.system() == "Darwin"):
        return "lib" + baseName + ".dylib"

def _buildFunction(call, args, result, errCheck = None):
    call.argtypes = args
    call.restype = result
    if (None != errCheck):
        call.errcheck = errCheck
    return call

def _isPython3():
    return sys.version_info[0] >= 3

def _createCharBuffer23(size):
    if (_isPython3()):
        return create_string_buffer(b'\000' * size)
    return create_string_buffer('\000' * size)

def _convertToString23(binaryString):
    if (_isPython3()):
        return binaryString.decode("utf-8")
    return binaryString

def _toString23(string):
    if (_isPython3()):
        return string.encode("utf-8")
    return string

def _checkError(result, func, arguments):
    if result != 0:
        error = _createCharBuffer23(BL_IF_BUFFER_SIZE)
        _CeGetLastErrorText(result, error)
        raise Exception(_convertToString23(error.value))

_library = cdll.LoadLibrary(_getLibraryName())

_BlGetLastErrorText = _buildFunction(_library.BlGetLastErrorText, [ c_int, POINTER(c_char) ], None)
_BlGetVersion = _buildFunction(_library.BlGetVersion, [POINTER(c_int), POINTER(c_int), POINTER(c_int)], c_int, _checkError)
_BlInit = _buildFunction(_library.BlInit, [], c_int, _checkError)
_BlEnumerate = _buildFunction(_library.BlEnumerate, [c_int, c_int, POINTER(c_int)], c_int, _checkError)
_BlCleanup = _buildFunction(_library.BlCleanup, [], c_int, _checkError)
_BlSetLogLevel = _buildFunction(_library.BlSetLogLevel, [c_int], c_int, _checkError)
_BlOpen = _buildFunction(_library.BlOpen, [c_int, POINTER(c_int)], c_int, _checkError)
_BlClose = _buildFunction(_library.BlClose, [c_int], c_int, _checkError)
_BlReadRegister = _buildFunction(_library.BlReadRegister, [c_int, c_uint, POINTER(c_uint)], c_int, _checkError)
_BlWriteRegister = _buildFunction(_library.BlWriteRegister, [c_int, c_uint, c_uint], c_int, _checkError)
_BlReadBlock = _buildFunction(_library.BlReadBlock, [c_int, c_uint, c_void_p, c_int, c_int], c_int, _checkError)
_BlWriteBlock = _buildFunction(_library.BlWriteBlock, [c_int, c_uint, c_void_p, c_int, c_int], c_int, _checkError)
_BlResetFpga = _buildFunction(_library.BlResetFpga, [c_int], c_int, _checkError)
_BlProgramFpgaFromBin = _buildFunction(_library.BlProgramFpgaFromBin, [c_int, c_char_p], c_int, _checkError)
_BlProgramFpgaFromMemory = _buildFunction(_library.BlProgramFpgaFromMemory, [c_int, c_void_p, c_int, c_int], c_int, _checkError)
_BlSetTimeout = _buildFunction(_library.BlSetTimeout, [c_int, c_uint], c_int, _checkError)
_BlSetUserId = _buildFunction(_library.BlSetUserId, [c_int, POINTER(c_char)], c_int, _checkError)
_BlGetUserId = _buildFunction(_library.BlGetUserId, [c_int, POINTER(c_char)], c_int, _checkError)
_BlGetSerialNumber = _buildFunction(_library.BlGetSerialNumber, [c_int, POINTER(c_char)], c_int, _checkError)
_BlGetFirmwareVersion = _buildFunction(_library.BlGetFirmwareVersion, [c_int, POINTER(c_int), POINTER(c_int), POINTER(c_int)], c_int, _checkError)
_BlGetDerivateInfo = _buildFunction(_library.BlGetDerivateInfo, [c_int, POINTER(c_char)], c_int, _checkError)
_BlGetDerivateId = _buildFunction(_library.BlGetDerivateId, [c_int, POINTER(c_uint)], c_int, _checkError)

BL_IF_BUFFER_SIZE = 16384

_enumRound = 0

class DeviceInfo:
    """
    This type offers access to most of the device information. Some information
    is available after opening the device. An instance to this type can either
    be retrieved from EnumeratedDevice.getDeviceInfo() or Device.getDeviceInfo().
    """
    def __init__(self, vid, pid):
        self._handle = -1
        self._vid = vid
        self._pid = pid

    def getVid(self): return self._vid
    def getPid(self): return self._pid

    def setUserId(self, userId):
        """
        Set user ID. Writes the string to the device EEPROM.
        This ID is intended to differentiate multiple devices of the same type connected
        to one host. The ID can contain up to 63 ASCII characters.
        """
        if self._handle == -1:
            raise Exception("Device not opened.")

        buf = _createCharBuffer23(BL_IF_BUFFER_SIZE)
        usid = _toString23(userId)
        for i in range(len(usid)):
            buf[i] = usid[i]
        _BlSetUserId(self._handle, buf)

    def getUserId(self):
        """
        Return user ID from device EEPROM. See explanation in
        setUserId(). "" if not accessible or in case of error.
        """
        if self._handle == -1:
            return ""

        try:
            dinfo = _createCharBuffer23(BL_IF_BUFFER_SIZE)
            _BlGetUserId(self._handle, dinfo)
            return _convertToString23(dinfo.value)
        except:
            return ""

    def getSerialNumber(self):
        """
        Return serial number. Not supported by all device
        types / instances. "" if not accessible or in case of error.
        """
        if self._handle == -1:
            return ""

        try:
            dinfo = _createCharBuffer23(BL_IF_BUFFER_SIZE)
            _BlGetSerialNumber(self._handle, dinfo)
            return _convertToString23(dinfo.value)
        except:
            return ""

    def getFirmwareVersion(self):
        """
        Return tuple of full version (major, minor, fix).
        """
        if self._handle == -1:
            return (0, 0, 0)

        try:
            major, minor, fix = c_int(), c_int(), c_int()
            _BlGetFirmwareVersion(self._handle, byref(major), byref(minor), byref(fix))
            return (major.value, minor.value, fix.value)
        except:
            return (0, 0, 0)

    def getDerivateInfo(self):
        """
        Return derivate info string. "" if not accessible or in case of error.
        Some devices are available in different versions (different FPGA's, different peripherals).
        The returned string contains a textual information about the specific
        derivate. For programmatically differentiation, getDerivateId() is better suited.
        """
        if self._handle == -1:
            return ""

        try:
            dinfo = _createCharBuffer23(BL_IF_BUFFER_SIZE)
            _BlGetDerivateInfo(self._handle, dinfo)
            return _convertToString23(dinfo.value)
        except:
            return ""

    def getDerivateId(self):
        """
        Return derivate id. 0xffffffff if not accessible or in case of error.
        This ID is unique for every derivate of a specific device type.
        Existing derivates can be found in the documentation of the specific device type.
        """
        if self._handle == -1:
            return 0xffffffff

        try:
            r = c_uint()
            _BlGetDerivateId(self._handle, byref(r))
            return r.value
        except:
            return 0xffffffff

class Device:
    """
    This type offers functionality to communicate with a single device. Instances can be
    retrieved from EnumeratedDevice.open() and represent an opened connection to a device.
    """

    # This flag can be used in readBlock() and writeBlock() to specify transfer to a
    # constant address. The opposite and *default* behavior is incremental address.
    # Attention: For historical reasons, this flag is interpreted inverse inside the FPGA interface ->
    # "address auto increment" is 0 if FlagConstAddress is specified.
    FlagConstAddress = 0x00000001

    def __verify32Bit(self, value):
        """
        Check if value is in range 0 <= value <= 0xffffffff
        """
        if(value < 0 or value > 0xffffffff):
            raise AttributeError("Argument out of range (must be 32bit unsigned).")

    def __init__(self, info, handle):
        """
        Called by the API, do not call yourself.
        """
        self._info = info
        self._info._handle = handle
        self._handle = handle

    def __enter__(self):
        """
        So it can be used by "with".
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        So it can be used by "with".
        """
        self.close()

    def close(self):
        """
        Free access to device. After this call the device is available in new enumerations and
        this instance of type Device can't be used anymore.
        """
        _BlClose(self._handle)

    def readRegister(self, reg):
        """
        Read 32 bit register value.
        @param reg Register in FPGA address space.
        @return Value of register.
        """
        self.__verify32Bit(reg)
        r = c_uint()
        _BlReadRegister(self._handle, reg, byref(r))
        return r.value

    def writeRegister(self, reg, value):
        """
        Write 32 bit register value.
        @param reg Register in FPGA address space.
        @param value Value to write to register.
        """
        self.__verify32Bit(reg)
        self.__verify32Bit(value)
        _BlWriteRegister(self._handle, reg, value)

    def readBlock(self, addr, buf, flags = 0):
        """
        Read data block.
        @param addr Start address in FPGA address space.
        @param buf Space to read to from FPGA address space, must be of type 'array.array'.
        @param flags I/O specific flags (see their individual documentation)
                - Device::FlagConstAddress
        """
        size = buf.itemsize * buf.buffer_info()[1]
        self.__verify32Bit(addr)
        self.__verify32Bit(size)
        _BlReadBlock(self._handle, addr, buf.buffer_info()[0], size, flags)

    def writeBlock(self, addr, buf, flags = 0):
        """
        Write data block.
        @param addr Start address in FPGA address space.
        @param buf Data to write to FPGA address space, must be of type 'array.array'.
        @param flags I/O specific flags (Device::FlagConstAddress).
        """
        size = buf.itemsize * buf.buffer_info()[1]
        self.__verify32Bit(addr)
        self.__verify32Bit(size)
        _BlWriteBlock(self._handle, addr, buf.buffer_info()[0], size, flags)

    def resetFpga(self):
        """
        Pulse FPGA reset line (the I/O that is chosen to be reset). Useful to bring the FPGA
        in a defined state. Implicitly done after FPGA programming.
        """
        _BlResetFpga(self._handle)

    def programFpgaFromBin(self, fileName):
        """
        Configure FPGA with given design in file. Loads a .bin from as specified by
        fileName and programs the FPGA using this design. Subsequently pulses the reset line.
        @param fileName File which contains design file data.
        """
        _BlProgramFpgaFromBin(self._handle, _toString23(fileName))

    def programFpgaFromMemory(self, buf, compressed):
        """
        Configure FPGA with given design in memory. Useful when compiling FPGA designs into
        the executable. BoardManager creates sources that can be used this way.
        @param buf Design, must be of type 'array.array'.
        @param compressed True if data is compressed, otherwise False.
        """
        size = buf.itemsize * buf.buffer_info()[1]
        _BlProgramFpgaFromMemory(self._handle, buf.buffer_info()[0], size, 1 if compressed else 0)

    def setTimeout(self, timeout):
        """
        Configure timeout until a pending transfer is stopped and defined as failed.
        This timeout is used for all transfers to and from device.
        @param timeout Timespan in milliseconds.
        """
        _BlSetTimeout(self._handle, timeout)

    def getDeviceInfo(self):
        """
        Get general device info.
        """
        return self._info

class EnumeratedDevice:
    """
    Entry returned by enumeration, valid only between two
    enumerations. Call its open() method the get the
    real device handle in form of an instance of type Device.
    """
    def __init__(self, info, eid, erid):
        self._info = info
        self._eid = eid
        self._erid = erid

    def open(self):
        """
        Open this instance, returns object to use for further communication.
        An opened device isn't reported in any subsequent enumeration (system wide), until
        it's closed.
        """
        if (self._erid != _enumRound):
            raise Exception("Invalid enumeration round.")

        handle = c_int()
        _BlOpen(self._eid, byref(handle))

        return Device(self._info, handle.value)

    def getDeviceInfo(self):
        """
        Get general device info.
        """
        return self._info

def getVersion():
    """
    Return tuple of full version (major, minor, fix).
    """
    major, minor, fix = c_int(), c_int(), c_int()
    _BlGetVersion(byref(major), byref(minor), byref(fix))
    return (major.value, minor.value, fix.value)

def init():
    """
    The API is not usable before this function is called.
    """
    _BlInit()

def enumerate(vid, pid):
    """
    Search for devices by the given VID/PID.
    @param vid Vendor-ID.
    @param pid Product-ID.
    @return List of detected devices. Instances are valid until the next call to enumerate.
    """
    ++_enumRound

    count = c_int()
    _BlEnumerate(vid, pid, byref(count))
    r = []

    for i in range(count.value):
        info = DeviceInfo(vid, pid)
        r.append(EnumeratedDevice(info, i, _enumRound))

    return r

def cleanup():
    """
    Free API resources, closes all devices and unloads all modules. To reuse the API again,
    init() must be called again.
    """
    _BlCleanup()

def setLogLevel(level):
    """
    Set logging level (Off = 0, Error = 1, Warning = 2, Debug = 3), default is 2.
    """
    _BlSetLogLevel(level)
