/* 
  CESYS Software License - Version 1.0 - January 1st, 2016

  Permission is hereby granted, free of charge, to any person or organization
  obtaining a copy of the software and accompanying documentation covered by
  this license (the "Software") to use, reproduce, display, distribute,
  execute, and transmit the Software, and to prepare derivative works of the
  Software, and to permit third-parties to whom the Software is furnished to
  do so, all subject to the following:

  The software must only be used to operate hardware manufactured by 
  CESYS GmbH, Herzogenaurach.

  The copyright notices in the Software and this entire statement, including
  the above license grant, this restriction and the following disclaimer,
  must be included in all copies of the Software, in whole or in part, and
  all derivative works of the Software, unless such copies or derivative
  works are solely in the form of machine-executable object code generated by
  a source language processor.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
  SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
  FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
  
  Written by Thomas Hoppe <thomas.hoppe@cesys.com>, 2018
*/  

package com.cesys.beastlink;


import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import static com.cesys.beastlink.SystemConnector.*;

/**
 * This type offers functionality to communicate with a single device. Instances can be
 * retrieved from EnumeratedDevice.open() and represent an opened connection to a device.
 */
public class Device implements AutoCloseable {
	private final DeviceInfo _info;
	private final int _handle;

	/**
	 * This flag can be used in readBlock() and writeBlock() to specify transfer to a
	 * constant address. The opposite and *default* behavior is incremental address.
	 * Attention: For historical reasons, this flag is interpreted inverse inside the FPGA interface ->
	 * "address auto increment" is 0 if FlagConstAddress is specified.
	 */
	public final static int FlagConstAddr = BL_FLAG_CONST_ADDR;

	/**
	 * Get general device info.
	 */
	public DeviceInfo getDeviceInfo() {
		return _info;
	}

	Device(DeviceInfo info, int handle) {
		_info = info;
		_info.setHandle(handle);
		_handle = handle;
	}

	/**
	 * Free access to device. After this call the device is available in new enumerations and
	 * this instance of type Device can't be used anymore.
	 */
	@Override
	public void close() {
		try {
			checkError(BlClose(_handle));
		} catch (IOException ex) {
			ex.printStackTrace();
		}
	}

	/**
	 * Read 32 bit register value.
	 * @param reg Register in FPGA address space.
	 * @return Value of register.
	 * @throws IOException
	 */
	public long readRegister(long reg) throws IOException {
		int[] r = new int[1];
		checkError(BlReadRegister(_handle, toUint32(reg), r));
		return fromUInt32(r[0]);
	}

	/**
	 * Write 32 bit register value.
	 * @param reg Register in FPGA address space.
	 * @param value Value to write to register.
	 * @throws IOException
	 */
	public void writeRegister(long reg, long value) throws IOException {
		checkError(BlWriteRegister(	_handle, toUint32(reg), toUint32(value)));
	}

	/**
	 * Read data block.
	 * @param addr Start address in FPGA address space.
	 * @param buffer Space to read to from FPGA address space.
	 * @throws IOException
	 */
	public void readBlock(long addr, byte[] buffer) throws IOException {
		readBlock(addr, buffer, 0);
	}

	/**
	 * Read data block.
	 * @param addr Start address in FPGA address space.
	 * @param buffer Space to read to from FPGA address space.
	 * @param flags I/O specific flags (see their individual documentation)
					- Device.FlagConstAddress
	 * @throws IOException
	 */
	public void readBlock(long addr, byte[] buffer, int flags) throws IOException {
		checkError(BlReadBlock(_handle, toUint32(addr), buffer, buffer.length, flags));
	}
	
	/**
	 * Read data block.
	 * @param addr Start address in FPGA address space.
	 * @param bytes Amount of bytes to read.
	 * @throws IOException
	 */
	public byte[] readBlock(long addr, int bytes) throws IOException {
		byte[] r = new byte[bytes];
		readBlock(addr, r, 0);
		return r;
	}

	/**
	 * Read data block.
	 * @param addr Start address in FPGA address space.
	 * @param bytes Amount of bytes to read.
	 * @param flags I/O specific flags (see their individual documentation)
					- Device.FlagConstAddress
	 * @throws IOException
	 */
	public byte[] readBlock(long addr, int bytes, int flags) throws IOException {
		byte[] r = new byte[bytes];
		readBlock(addr, r, flags);
		return r;
	}	

	/**
	 * Write data block.
	 * @param addr Start address in FPGA address space.
	 * @param buffer Data to write to FPGA address space.
	 * @throws IOException
	 */
	public void writeBlock(long addr, byte[] buffer) throws IOException {
		writeBlock(addr, buffer, 0);
	}

	/**
	 * Read data block.
	 * @param addr Start address in FPGA address space.
	 * @param buffer Data to write to FPGA address space.
	 * @param flags I/O specific flags (see their individual documentation)
					- Device.FlagConstAddress
	 * @throws IOException
	 */
	public void writeBlock(long addr, byte[] buffer, int flags) throws IOException {
		checkError(BlWriteBlock(_handle, toUint32(addr), buffer, buffer.length, flags));
	}

	/**
	 * Pulse FPGA reset line (the I/O that is chosen to be reset). Useful to bring the FPGA
	 * in a defined state. Implicitly done after FPGA programming.
	 * @throws IOException
	 */
	public void resetFpga() throws IOException {
		checkError(BlResetFpga(_handle));
	}

	/**
	 * Configure FPGA with given design in file. Loads a .bin from as specified by
	 * fileName and programs the FPGA using this design. Subsequently pulses the reset line.
	 * @param fileName File which contains design file data.
	 * @throws IOException
	 */
	public void programFpgaFromBin(String fileName) throws IOException {
		checkError(BlProgramFpgaFromBin(_handle, fileName));
	}

	/**
	 * Configure FPGA with given design in memory. Useful when compiling FPGA designs into
	 * the executable. BoardManager creates sources that can be used this way.
	 * Subsequently pulses the reset line.
	 * @param buffer Design buffer.
	 * @param compressed True if resource is compressed.
	 * @throws IOException
	 */
	public void programFpgaFromMemory(byte[] buffer, boolean compressed) throws IOException {
		checkError(BlProgramFpgaFromMemory(_handle, buffer, buffer.length, compressed ? 1 : 0));
	}

	/**
	 * JVM specific function that let the design load from class path.
	 * The .bin file is then loaded from stream clazz.getResourceAsString(path).
	 * @param clazz See java.lang.Class.getResourceAsStream().
	 * @param path See java.lang.Class.getResourceAsStream().
	 * @throws IOException
	 */
	public void programFpgaFromResource(Class<?> clazz, String path) throws IOException {
		// Read data from resource.
		try (InputStream is = clazz.getResourceAsStream(path)) {
			if (is == null)
				throw new IOException("Invalid resource specified.");

			// Read to ByteOutputStream.
			try (ByteArrayOutputStream bos = new ByteArrayOutputStream()) {
				final int bs = 16384;
				byte[] buf = new byte[bs];
				int len = 0;
				while ((len = is.read(buf, 0, bs)) != -1)
					bos.write(buf, 0, len);

				// This is the resource.
				byte[] design = bos.toByteArray();

				// The bytes must be reversed to use in programFpgaFromMemory().
				for (int i = 0; i < design.length; ++i)
					design[i] = (byte) ((Integer.reverse(design[i]) >> 24) & 0xff);

				// Finally program the design.
				programFpgaFromMemory(design, false);
			}
		}
	}

	/**
	 * Configure timeout until a pending transfer is stopped and defined as failed.
	 * This timeout is used for all transfers to and from device.
	 * @param timeout Time span in milliseconds.
	 * @throws IOException
	 */
	public void setTimeout(int timeout) throws IOException {
		checkError(BlSetTimeout(_handle, timeout));
	}
}
