/* 
  CESYS Software License - Version 1.0 - January 1st, 2016

  Permission is hereby granted, free of charge, to any person or organization
  obtaining a copy of the software and accompanying documentation covered by
  this license (the "Software") to use, reproduce, display, distribute,
  execute, and transmit the Software, and to prepare derivative works of the
  Software, and to permit third-parties to whom the Software is furnished to
  do so, all subject to the following:

  The software must only be used to operate hardware manufactured by 
  CESYS GmbH, Herzogenaurach.

  The copyright notices in the Software and this entire statement, including
  the above license grant, this restriction and the following disclaimer,
  must be included in all copies of the Software, in whole or in part, and
  all derivative works of the Software, unless such copies or derivative
  works are solely in the form of machine-executable object code generated by
  a source language processor.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
  SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
  FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
  
  Written by Thomas Hoppe <thomas.hoppe@cesys.com>, 2018
*/  

package com.cesys.beastlink.example;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.stream.IntStream;

import com.cesys.beastlink.Device;
import com.cesys.beastlink.DeviceInfo;
import com.cesys.beastlink.EnumeratedDevice;
import com.cesys.beastlink.LibraryInterface;

public class Example {
	// We use VID and PID for EFM-03 devices.
	private final int VID = 0x10f8;
	private final int PID = 0xc583;
	
	/**
	 * Do a device/FPGA design example specific test for each device that ships
	 * with an example design.
	 *
	 * @throws IOException
	 */
	private void deviceTest() throws IOException {
		// Loop over all devices.
		for (EnumeratedDevice ed : LibraryInterface.enumerate(VID, PID)) {
			// Open the device.
			try (Device d = ed.open()) {
				// This is the example design by Cesys, the design is read from 
				// the specified file and programmed to the FPGA on the device (must be .bin format).
				// Optionally, programFpgaFromMemory() can
				// be used to download a compiled in resource (generated by tools).
				// resetFpga() is explicitly invoked from within this call.
				d.programFpgaFromBin("./beastlink_exdes_top.bin");
				
				// The example maps 16K Block RAM onto the AXI-4 Bus, starting at address 0x00000000.
				final long BlockRAMAddr = 0x00000000L;
				final int BlockRAMSize = 16 * 1024;
				final int HalfRAMSize = BlockRAMSize / 2;
				
				// Create 16kb block containing counting 32 bit values.
				byte[] data = new byte[BlockRAMSize];
				ByteBuffer bb = ByteBuffer.wrap(data);
				IntStream.range(0, BlockRAMSize / 4).forEachOrdered(bb::putInt);
				
				// Write to Block RAM.
				d.writeBlock(BlockRAMAddr, data);
				
				// Read back, starting at 8k.
				byte[] rbdata = d.readBlock(BlockRAMAddr + HalfRAMSize, HalfRAMSize);
				
				// Compare second half of written data to read back data.
				byte[] shalf = Arrays.copyOfRange(data, HalfRAMSize, BlockRAMSize);
				
				// Write compare result.
				System.out.print("Sequences are " + (Arrays.equals(rbdata, shalf) ? "equal" : "not equal") + ".");				
			}
		}
	}

	/**
	 * Enumerate all known devices and write information to stdout.
	 *
	 * @throws IOException
	 */
	private void infoTest() throws IOException {
		// Loop over all devices.
		for (EnumeratedDevice ed : LibraryInterface.enumerate(VID, PID)) {
			// EnumeratedDevice.open() returns the opened instance to the device. 
			try (Device d = ed.open()) {
				// This structure contains all information about the device.
				// Not all devices have identical information. Devices, which
				// don't have a property return a default value, no error is thrown.
				DeviceInfo i = d.getDeviceInfo();
	
				// Write all informations to stdout, values are self explaining.
				System.out.println("Device found:");
				System.out.println("  Vid             : " + String.format("%04x", i.getVid()));
				System.out.println("  Pid             : " + String.format("%04x", i.getPid()));
				System.out.println("  UserId          : " + i.getUserId());
				System.out.println("  SerialNumber    : " + i.getSerialNumber());
				System.out.println("  FirmwareVersion : " + i.getFirmwareVersion());
				System.out.println("  DerivateInfo    : " + i.getDerivateInfo());
				System.out.println("  DerivateId      : 0x" + String.format("%08x", i.getDerivateId()));
			}
		}
	}

	/**
	 * Application entry point.
	 */
	private void run() {
		// Any error is reported by IOException, catch it globally.
		try {
			// Initialize BeastLink before using it.
			LibraryInterface.init();

			// Output BeastLink version to stdout.
			System.out.println("BeastLink version : " + LibraryInterface.getVersion());

			// Set log level to log up to warning level (default).
			// Valid log levels are:
			// - 0x00: Off
			// - 0x01: Error
			// - 0x02: Warning
			// - 0x03: Debug
			LibraryInterface.setLogLevel(0x2);

			// Write all informations about all known connected
			// devices to stdout.
			infoTest();

			// Do a device specific test (with devices that have an
			// example FPGA design).
			deviceTest();

			// Cleanup internal structures. This closes all
			// opened resources, no further use of the API
			// is possible until init() is called again.
			LibraryInterface.cleanup();
		} catch (IOException ex) {
			System.out.println("Error: " + ex.getMessage());
			ex.printStackTrace();
		}
	}

	public static void main(String[] args) {
		new Example().run();
	}
}
